-------------------------- MODULE MCMSPastry ---------------------------
\* This is the model checking module, only used for model checking
(* The hierarchy of the modules are
 Ring (The basic parameters and operations for ring calculation)
   \subset LS (The leaf set utilities)
      \subset RT (The routing table utilities)
         \subset Msg (The messages definitions)
            \subset InitialStates (The initial state and TypeInvariant)
               \subset Actions (All the actions)
                 \subset MSPastry (The specification of MSPastry including propertie
                   \subset MCMSPastry (The model checking file)                  
*)
(* The hierarchy of the Proofs are
 ProofBasic  
  \subset ProofRingProp (The basic parameters and operations for ring calculation)
   \subset ProofLSetProp (The leaf set local properties of data structure)
    \subset ProofProp (reductions)
     \subset ProofType (The Proof of TypeInvariant)
      \subset ProofLSetInv (The proof of Invariant of Leaf Set properties)
*)
EXTENDS MSPastry 
------------------------------------------------------------------------
\* Demo Assignment
MCB == 1 \* base is 2^1 = 2
MCM == 2 \* ID space is 2^2 = 4
MCL == 1 \* leaf set length is 1, left one node, right one node
MCA == {1}
MCI == {1, 2, 3, 4}
LightNext == \E i, j \in I:
      \/ RouteJReq(i, j) 
      \/ RecProbe(i)
      \/ RecPRply(i) 
      \/ RecJReq(i)
      \/ RecJReply(i)
      \/ RecLReq(i)
      \/ RecLReply(i)
      \/ RequestLease(i)
      \/ Deliver(i, j)
      \/ RouteLookup(i, j) 

ND_Next == \E i, j \in I:
      \*\/ Deliver(i, j)
      \*\/ RouteLookup(i, j)  
      \/ RecProbe(i)
      \/ RecPRply(i)      
      \/ RecLReq(i)
      \/ RecLReply(i)       
      \/ LeaseExp(i, j)
      \/ GrantExp(i, j)
      \*\/ NodeLeft(i)      
      \/ SuspectFaulty(i, j)
      \/ TimeOut(i, j)
      \/ RequestLease(i)
      \*\/ MsgLost
      \*\/ LSRepair(i)      
      \*\/ Recover(i)
      \*\/ Restart(i)
      \*\/ ResignNode(i)   
JND_Next == \E i, j \in I:
      \*\/ Deliver(i, j)
      \*\/ RouteLookup(i, j)  
      \/ RecJReq(i)
      \/ RecJReply(i)  
      \/ RouteJReq(i, j)
      \/ RecProbe(i)
      \/ RecPRply(i)      
      \/ RecLReq(i)
      \/ RecLReply(i)       
      \/ LeaseExp(i, j)
      \/ GrantExp(i, j)
      \*\/ NodeLeft(i)      
      \/ SuspectFaulty(i, j)
      \/ TimeOut(i, j)
      \/ RequestLease(i)     

-------------------------------------------------------
\* The following are liveness properties and their specification configurations

(*Successful join*)
\* pure join
JoinNext == \/ Join(2, 1)
            \/ LightNext
SpecSuccessJoin == 
/\ Init 
/\ [][JoinNext]_vars
-----------------
\* with faulty recovery
JoinNextAll == \/ Join(2, 1)
               \/ JND_Next
SpecSuccessJoinAll == 
/\ Init 
/\ [][JoinNextAll]_vars
----------------------
\*property to refute
NeverLease == [](\A m \in receivedMsgs: m.mreq.type # "LeaseRequest")
NeverJoin == [](\A j \in I \ A: status[j]#"ready") 
-------------------------------------------------------
(*Successful Concjoin*)
MCCJA == {1, 4}
---------------------------
\* pure join
CJNext ==  \/ Join(2, 1)
           \/ Join(3, 4)    
           \/ LightNext
SpecConcJoin == 
/\ Init 
/\ [][CJNext]_vars           
---------------------------      
\* with faulty recovery    
CJNextAll ==   \/ Join(2, 1)
               \/ Join(3, 4)         
               \/ JND_Next
SpecConcJoinAll == 
/\ Init 
/\ [][CJNextAll]_vars    
----------------------
\*property to refute  
ConcurrentJoin == []~(status[1] = "ready" 
                          /\status[2] = "ready"
                          /\status[3]= "ready" 
                          /\status[4]= "ready"
                          )
--------------------------------------------------------    
(*Node departures and ls repair*)

MCA_ND == {1, 2, 3}
----------------------------------
\* pure faulty recovery
NextND == \/ NodeLeft(1)
          \/ ND_Next
SpecND == /\ Init 
          /\ [][NextND]_vars
----------------------
\*property to refute 
\* This part checks if the a neighbor node could ever declare a node dead
\* when it has left the network
\* NeverDeclareDead ==  [][~\E i, j \in I: DeclareDead(i, j)]_vars
\* This part checks if the network could recover to a stable status
\* after node departure 
NeverStableAfterRepair ==[]~(/\ status[1] = "dead"
                        /\ status[2] = "ready"
                        /\ status[3] = "ready"
                        \*/\ 95 \in ReadyOK
                        /\ ~ 1 \in GetLSetContent(lset[2]))  

-------------------------------------------------------
\* Successful recovery

MCA_ND_CJ == {1, 4}
CJNDNext ==  \/ Join(3, 4)
             \/ Join(2, 1)
             \/ NodeLeft(1)          
             \/ JND_Next             
SpecCcJoinND == 
/\ Init 
/\ [][CJNDNext]_vars
NeverConcurrentJoin == []~(status[4] = "ready" 
                          /\status[1] = "ready"
                          /\status[2]= "ready" 
                          /\status[3]= "ready"
                          )
NeverStableAfterND ==  []~(/\ status[1] = "dead"
                        /\ status[2] = "ready"
                        /\ status[3] = "ready"
                        /\ status[4] = "ready"
                        /\ ~ 1 \in GetLSetContent(lset[4]))                                                                          

=======================================================================
