-------------------------- MODULE MSPastry ---------------------------
\* This is the specification of MSPastry
(* The hierarchy of the modules are
 Ring (The basic parameters and operations for ring calculation)
   \subset LS (The leaf set utilities)
      \subset RT (The routing table utilities)
         \subset Msg (The messages definitions)
            \subset InitialStates (The initial state and TypeInvariant)
               \subset Actions (All the actions)
                 \subset MSPastry (The specification of MSPastry including propertie
                   \subset MCMSPastry (The model checking file) 
*)
\* This version is based on the paper Castro04, Haeber06 
\* and the source codes from www.FreePastry.com
EXTENDS Actions 
----------------------------------------------------------------------
\* here are all the global variables
vars == <<receivedMsgs, status, lset, probing, failed, lease, grant, toj>>


\* here are all possible actions to be applied for next step
Next == \E i, j \in I:
      \/ RouteLookup(i, j) 
      \/ RouteJReq(i, j)  
      \/ Deliver(i, j)
      \/ RecProbe(i)
      \/ RecPRply(i) 
      \/ RecJReq(i)
      \/ RecJReply(i)
      \/ RequestLease(i)
      \/ RecLReq(i)
      \/ RecLReply(i)
      \*\/ SendPeriodicBLS(i)
      \* followed are application (interface) actions:
      \/ Lookup(i, j)
      \/ Join(i, j)
      \/ MsgLost
      \/ LeaseExp(i, j)
      \/ GrantExp(i, j)
      \/ NodeLeft(i)
      \/ SuspectFaulty(i, j)
      \/ TimeOut(i, j)
      \*\/ LSRepair(i)
      \*\/ Recover(i)
      \*\/ Restart(i)
      \*\/ ResignNode(i)
Spec == \* The general specification of MSPastry
/\ Init 
/\ [][Next]_vars
----------------------------------------------------------------------------------
\* Properties 
IRN == \A i \in I, r \in ReadyNodes: i # r 
   => CwDist(i, RightNeighbor(lset[i])) =< CwDist(i, r) 
NRI == \A i \in I, r \in ReadyNodes: i # r 
   => CwDist(LeftNeighbor(lset[i]), i) =< CwDist(r, i) 
   
GrantHistR == \A l, i, r \in I:
  /\ toj[l] = i
  /\ i # r
  /\ r \in grant[l]
  /\ ~ i \in ReadyNodes
  /\ CwDist(i, r) < CwDist(l, r)
  => CwDist(i, RightNeighbor(lset[i])) =< CwDist(i, r) 

GrantHistL == \A i, l, r \in I:
   /\ toj[r] = i
   /\ i # l
   /\ l \in grant[r]
   /\ ~ i \in ReadyNodes
   /\ CwDist(l, i) < CwDist(l, r)
   => CwDist(LeftNeighbor(lset[i]), i) =< CwDist(l, i)
   
GrantNeighborNew == \A k, i \in I: 
     /\ i \in grant[k]
     /\ i # k
   => 
   /\ lset[i] # EmptyLS(i)
   /\ lset[k] # EmptyLS(k)
   /\ CwDist(k, RightNeighbor(lset[k])) =< CwDist(k, i)
   /\ CwDist(LeftNeighbor(lset[k]), k) =< CwDist(i, k)  
   /\ CwDist(i, RightNeighbor(lset[i])) =< CwDist(i, k)
   /\ CwDist(LeftNeighbor(lset[i]), i) =< CwDist(k, i)  
   
TojClosestL == \A r1, r2, i, k \in I: 
  /\ i # r1  
  /\ i # k    
  /\ toj[r1] = i
  /\ RightNeighbor(lset[r1]) = i  
  /\ ~ i \in ReadyNodes
  /\ toj[r2] = k
  /\ r2 # k
  => CwDist(r1, i) =< CwDist(k, i)

TojClosestR == \A r1, r2, i, k \in I: 
  /\ i # r1 
  /\ i # k   
  /\ toj[r1] = i
  /\ LeftNeighbor(lset[r1]) = i  
  /\ ~ i \in ReadyNodes
  /\ toj[r2] = k
  /\ r2 # k
  => CwDist(i, r1) =< CwDist(i, k)
  
         
SemToj == 
   \A r, i \in I: 
   /\ toj[r] = i
   /\ r # i
   /\ ~ i \in ReadyNodes
   => \/ RightNeighbor(lset[r]) = i 
      \/ LeftNeighbor(lset[r]) = i    
 
TojNoReady == \A i \in I: 
     /\ ~i \in ReadyNodes 
     /\ lset[i] # EmptyLS(i)
    => \E r \in ReadyNodes: toj[r] = i
      
TojDualR == \A r, i \in I:
           /\ r # i
           /\ toj[r] = i
           /\ ~ i \in ReadyNodes
           /\ RightNeighbor(lset[r]) = i
           /\ lset[i] # EmptyLS(i)
           => LeftNeighbor(lset[i]) = r
TojDualL == \A r, i \in I:
           /\ r # i
           /\ toj[r] = i
           /\ ~i \in ReadyNodes
           /\ LeftNeighbor(lset[r]) = i
           /\ lset[i] # EmptyLS(i)
           => RightNeighbor(lset[i]) = r            
TojReady == \A r \in I:
     toj[r] # r
     => r \in ReadyNodes     
TojInjective == \A i, j, k \in I: 
         /\ toj[i] = k 
         /\ toj[j] = k 
         /\ i # k 
         /\ j # k
        => i = j  
LSnonEmptyToj == \A i, j \in I: 
               /\ i \in GetLSetContent(lset[j])
               /\ i # j
              => \/ lset[i] # EmptyLS(i)
                 \/ toj[j] = i         

CoverNoToj == \A l, i, r \in I:
         /\ Covers(lset[r], i)
         /\ r \in grant[l]
         /\ toj[l] = i
         /\ i # r
         /\ r # l
         => FALSE
            
CoverNoNode == \A i \in ReadyNodes, k, j \in I:
               /\ Covers(lset[i], k) 
               /\ i # k              
               /\ j # k
               => toj[j] # k    
               
JoinNC == \A m \in receivedMsgs, r \in ReadyNodes: 
  /\ m.mreq.type = "JoinReply"
  /\ m.mreq.lset.node # r
  => /\ ( RightNeighbor(AddToLSet(GetLSetContent(m.mreq.lset), 
          EmptyLS(m.destination))) = m.mreq.lset.node 
        => CwDist(LeftNeighbor(m.mreq.lset), m.mreq.lset.node) 
             =< CwDist(r, m.mreq.lset.node))
     /\ ( LeftNeighbor(AddToLSet(GetLSetContent(m.mreq.lset), 
           EmptyLS(m.destination))) = m.mreq.lset.node
          => CwDist(m.mreq.lset.node, RightNeighbor(m.mreq.lset)) 
              =< CwDist(m.mreq.lset.node, r))    
JoinGrant == \A m \in receivedMsgs, g \in I:
  /\ m.mreq.type = "JoinReply"
  /\ g \in grant[m.mreq.lset.node]
  /\ g # m.mreq.lset.node
  => /\( RightNeighbor(AddToLSet(GetLSetContent(m.mreq.lset), 
          EmptyLS(m.destination))) = m.mreq.lset.node 
          => CwDist(LeftNeighbor(m.mreq.lset), m.mreq.lset.node)  
             =< CwDist(g, m.mreq.lset.node))
     /\ ( LeftNeighbor(AddToLSet(GetLSetContent(m.mreq.lset), 
           EmptyLS(m.destination))) = m.mreq.lset.node
          => CwDist(m.mreq.lset.node, RightNeighbor(m.mreq.lset)) 
             =< CwDist(m.mreq.lset.node, g))         
JoinNeighborR == \A m \in receivedMsgs, ls \in LSet: 
  /\ m.mreq.type = "JoinReply"
  /\ ls = AddToLSet(GetLSetContent(m.mreq.lset), EmptyLS(m.destination))
  /\ RightNeighbor(ls) = m.mreq.lset.node 
 => LeftNeighbor(lset[m.mreq.lset.node]) = m.destination

JoinNeighborL == \A m \in receivedMsgs, ls \in LSet: 
  /\ m.mreq.type = "JoinReply"
  /\ ls = AddToLSet(GetLSetContent(m.mreq.lset), EmptyLS(m.destination))
  /\ LeftNeighbor(ls) = m.mreq.lset.node 
 => RightNeighbor(lset[m.mreq.lset.node]) = m.destination
JoinSingleNode == \A m \in receivedMsgs: 
  /\ m.mreq.type = "JoinReply"
  /\ LeftNeighbor(lset[m.mreq.lset.node]) = RightNeighbor(lset[m.mreq.lset.node])
  => m.mreq.lset = EmptyLS(m.mreq.lset.node)     
                
SemBLSReady == 
     \A m \in receivedMsgs:
           /\ m.mreq.type = "LeaseReply"
           /\ toj[m.destination] = m.mreq.lset.node         
         => m.mreq.lset.node \in ReadyNodes

SemBLSnonEmpty == \A m \in receivedMsgs: 
     m.mreq.type = "LeaseReply"
     => /\ lset[m.mreq.lset.node] # EmptyLS(m.mreq.lset.node)
        /\ lset[m.destination] # EmptyLS(m.destination)         
SemLReq == \A m \in receivedMsgs: 
          m.mreq.type = "LeaseRequest" 
         => /\ toj[m.mreq.node] # m.destination
            /\ m.mreq.node \in ReadyOK         
            /\ m.mreq.node # m.destination
            /\ lset[m.mreq.node] # EmptyLS(m.mreq.node)
            /\ lset[m.destination] # EmptyLS(m.destination)
            /\ CwDist(m.mreq.node, RightNeighbor(lset[m.mreq.node])) 
               =< CwDist(m.mreq.node, m.destination)
            /\ CwDist(LeftNeighbor(lset[m.mreq.node]), m.mreq.node) 
               =< CwDist(m.destination, m.mreq.node)    
                 


----------------------------------------------                   
 (*
LSetInv == \A i \in I: ~i \in (lset[i].left) /\ ~i \in (lset[i].right) 
Stable == \*A node is stable ("ok" or "ready") iff
\*its leaf set is complete or there are not enough nodes to complete any leaf set
\* This is not used for Theorem Proving and not proved, 
\* but MC hasn't found Counter Example till depth 21
  \/ Cardinality(ReadyOK) =< L
  \/ \A i \in ReadyOK : IsComplete(lset[i])
*)
\* no other node could be closer to a node than its neighbor 
NeighborClosest == 
  \/ Cardinality(ReadyNodes) =< 1
  \/ \A x, y \in ReadyNodes: 
     ~(x = y) 
     => 
       /\ CwDist(LeftNeighbor(lset[x]), x)=< CwDist(y, x)
       /\ CwDist(x, RightNeighbor(lset[x])) =< CwDist(x, y)
                 
HalfNeighbor == 
  \/ \A k \in ReadyOK : 
      /\ RightNeighbor(lset[k]) # k 
      /\ LeftNeighbor(lset[k]) # k
  \/ \E k \in ReadyOK:
     /\ ReadyOK = {k} 
     /\ LeftNeighbor(lset[k]) = k 
     /\ RightNeighbor(lset[k]) = k
     /\ \A w \in WaitNodes: GetLSetContent(lset[w]) \in SUBSET {k, w} 
     /\ ~\E ms \in receivedMsgs: ms.mreq.type = "ProbeReply" 
     /\ ~\E mk \in receivedMsgs: /\ mk.mreq.type = "Probe" 
                                 /\ mk.destination # k
     /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                               => GetLSetContent(mj.mreq.lset) = {k}
     /\ ~\E mb \in receivedMsgs: mb.mreq.type = "LeaseReply"
HalfNeighborReduction == 
  \/ \A k \in ReadyOK : 
      /\ RightNeighbor(lset[k]) # k 
      /\ LeftNeighbor(lset[k]) # k
  \/ \E k \in ReadyOK:
     /\ ReadyOK = {k} 
     /\ LeftNeighbor(lset[k]) = k 
     /\ RightNeighbor(lset[k]) = k
--------------------------------------------------------------------------------
(*
NeighborClosestRight == 
\* This does not hold by First Join Last ProbeReply case
\* let node a, b, c, d \in I lies in the written order clockwise on a ring.
\* d is ready and b joins the ring by d
\* After d replies b with itself in the leaf set, b updates its leaf set with d and probe d
\* Before d received the probe, a and c joins the ring by d and completes the 
\* consistent join phrase and recognize each other as direct neighbors. 
\* Then a and c get the grands from c and a and d to become ready, while b is still waiting for the probe reply.
\* Now b's neighbor is d but a and c are closer to it and ready.
\* Hence, b's status is still not jet OK, because even it has d to fill the complete leaf set condition,
\* it still has outgoing probing. 
\A k \in ReadyNodes, i, j \in I: 
 /\ i # k
 /\ i # j
 /\ j # k
 /\ j = RightNeighbor(lset[i])
 /\ CwDist(i, k) < CwDist(i, j)
 => k \in GetLSetContent(lset[i])
NeighborClosestLeft == 
\A k, i \in ReadyOK: i # k => CwDist(LeftNeighbor(lset[i]), i) =< CwDist(k, i)

*)
------------------------------------------------------------------   
GrantOK == \A i, j \in I: j \in grant[i]/\ i # j 
            => i \in ReadyOK /\ j \in ReadyOK                               
SingleReadyNode == \A i \in I: 
    ReadyNodes = {i} => LeftNeighbor(lset[i]) = RightNeighbor(lset[i])
LSnonEmpty == \A i, j \in I: 
               /\ i \in GetLSetContent(lset[j]) \ {toj[j]}
               /\ i # j
               \*/\ ~i \in ReadyNodes
               => lset[i] # EmptyLS(i)
LeaseGrant == \A i, n \in I:
            n \in lease[i] => i \in grant[n]

SingleReadyNodeToj == \A k \in I, r \in I:
     /\ ReadyNodes = {r}
     /\ lset[k] # EmptyLS(k)
     /\ r # k
    => toj[r] = k
AtleastOneReady == ReadyNodes # {}    
---------------------------------------------------------------------         
SemEmptyLSetSingleNode == \A m \in receivedMsgs:
                    /\ m.mreq.type = "JoinReply"
                    => 
                    (  \/ m.mreq.lset.node  = LeftNeighbor(m.mreq.lset)
                       \/ m.mreq.lset.node  = RightNeighbor(m.mreq.lset)
                       
                       <=> m.mreq.lset = EmptyLS(m.mreq.lset.node))
SemSingleReadyJoin == \A m \in receivedMsgs:
                    /\ m.mreq.type = "JoinReply"
                    /\ m.mreq.lset = EmptyLS(m.mreq.lset.node)
                    => /\ lset[m.mreq.lset.node] = m.mreq.lset
                       /\ ReadyNodes = {m.mreq.lset.node}
SemDead == \A k \in I: status[k] = "dead" =>  /\ lset[k] = EmptyLS(k)
                                              /\ probing[k] = {}   
SemProbePRply == \A m \in receivedMsgs: 
                  \/ m.mreq.type = "ProbeReply"
                  \/ m.mreq.type = "Probe"
                  => 
                     /\ m.mreq.node = m.mreq.lset.node
                     /\ m.mreq.node # m.destination
                     /\ ~ m.destination \in m.mreq.failed 
                     /\ lset[m.mreq.node] # EmptyLS(m.mreq.node)   
SemPRply ==    \A m \in receivedMsgs:
                       m.mreq.type = "ProbeReply"
                       => /\ lset[m.destination] # EmptyLS(m.destination)
                          /\ m.mreq.node \in probing[m.destination]  
SemProbe ==    \A m \in receivedMsgs:
                       m.mreq.type = "Probe"
                       => m.destination \in probing[m.mreq.node]   

PrbMsgDual == \A mp, mr \in receivedMsgs:
   /\ mr.mreq.type  = "ProbeReply"
   /\ mp.mreq.type = "Probe"
   => \/ mp.mreq.node # mr.destination
      \/ mr.mreq.node # mp.destination
OnePRply == \A m,mm \in receivedMsgs:
   /\ m.mreq.type = "ProbeReply"
   /\ mm.mreq.type = "ProbeReply"
   /\ m.destination = mm.destination
   /\ m.mreq.node = mm.mreq.node
   => m = mm
OneProbe == \A m,mm \in receivedMsgs:
   /\ m.mreq.type = "Probe"
   /\ mm.mreq.type = "Probe"
   /\ m.destination = mm.destination
   /\ m.mreq.node = mm.mreq.node
   => m = mm                                                                                                             
SemJoinNeighbor == \A ls \in LSet, m \in receivedMsgs:
                 /\ m.mreq.type = "JoinReply"
                 /\ ls = AddToLSet(GetLSetContent(m.mreq.lset), EmptyLS(m.destination))
                 =>  \/ /\ RightNeighbor(ls) = m.mreq.lset.node
                        /\ LeftNeighbor(ls) = LeftNeighbor(m.mreq.lset)
                     \/ /\ LeftNeighbor(ls) = m.mreq.lset.node 
                        /\ RightNeighbor(ls) = RightNeighbor(m.mreq.lset)
SemJoinLeafSet == \A m \in receivedMsgs:
                    m.mreq.type = "JoinReply"
                 => LET n == m.mreq.lset.node
                    IN 
                   /\ CwDist(LeftNeighbor(lset[n]), n) 
                        =< CwDist(LeftNeighbor(m.mreq.lset), n)
                   /\ CwDist(n, RightNeighbor(lset[n])) 
                        =< CwDist(n, RightNeighbor(m.mreq.lset))                
SemJRply == \A  m \in receivedMsgs: m.mreq.type = "JoinReply"
             =>  
                 /\ lset[m.destination] = EmptyLS(m.destination)
                 /\ toj[m.mreq.lset.node] = m.destination
                 /\ m.mreq.lset.node \in ReadyNodes
                 /\ m.destination \in WaitNodes     
                 /\ probing[m.destination] = {}            
                 \*/\ (LeftNeighbor(m.mreq.lset) # m.mreq.lset.node => 
                 \*   toj[LeftNeighbor(m.mreq.lset)] = LeftNeighbor(m.mreq.lset))
                 \*/\ (RightNeighbor(m.mreq.lset) # m.mreq.lset.node => 
                 \*    toj[RightNeighbor(m.mreq.lset)] = RightNeighbor(m.mreq.lset))
                 \*/\ RightNeighbor(lset[m.mreq.lset.node]) \in lease[m.mreq.lset.node]
                 \*/\ LeftNeighbor(lset[m.mreq.lset.node]) \in lease[m.mreq.lset.node]
                 \*/\ Covers(lset[m.mreq.lset.node], m.destination)
SemJRplyLSnonEmpty == \A m \in receivedMsgs: m.mreq.type = "JoinReply"
                        => \A k \in GetLSetContent(m.mreq.lset): 
                                  lset[k] # EmptyLS(k)  
SemBLS == 
        \A i, n \in I, m \in receivedMsgs:
              /\ m.mreq.type = "LeaseReply"
              /\ m.mreq.lset.node = n
              /\ m.destination = i
            =>
             /\ n \in ReadyOK
             /\ (m.mreq.grant =TRUE => i \in grant[n])
             /\ n # i             
SemJReq == \A m \in receivedMsgs: m.mreq.type  = "JoinRequest"
           => /\ m.mreq.node \in WaitNodes
              /\ lset[m.mreq.node] = EmptyLS(m.mreq.node)
              /\ probing[m.mreq.node] = {}
SemLease == \A i \in I: i \in lease[i]              
-----------------------------------------------------
OneJRply == \A m,mm \in receivedMsgs:
   /\ m.mreq.type = "JoinReply"
   /\ mm.mreq.type = "JoinReply"
   /\ m.destination = mm.destination
   => m = mm
OneJReq == \A m1, m2 \in receivedMsgs:
   /\ m1.mreq.type = "JoinRequest"
   /\ m2.mreq.type = "JoinRequest"
   /\ m1.mreq.node = m2.mreq.node
   => m1 = m2
JoinMsgDual == \A m, mm \in receivedMsgs: 
   /\ m.mreq.type  = "JoinReply"
   /\ mm.mreq.type = "JoinRequest"
   => mm.mreq.node # m.destination

-----------------------------------------------------
CorrectRouting == ~\E m \in receivedMsgs:m.mreq.type = "NoLegalRoute"          
-----------------------------------------------------
Invariant ==
 /\ IRN
 /\ NRI
 /\ GrantNeighborNew 
 /\ GrantHistL
 /\ GrantHistR
 /\ GrantOK  
 /\ CoverNoNode
 /\ CoverNoToj 
 /\ TojClosestL
 /\ TojClosestR
 /\ TojNoReady
 /\ SemToj
 /\ TojDualR
 /\ TojDualL
 /\ TojReady  
 /\ TojInjective 
 /\ LSnonEmptyToj
 /\ JoinNC  
 /\ JoinGrant
 /\ JoinNeighborR
 /\ JoinNeighborL
 /\ JoinSingleNode 
 /\ SemBLSnonEmpty
 /\ SemBLSReady  
 /\ SemLReq  
 /\ SemBLS
 /\ SemJRply 
 /\ SemJoinLeafSet 
 /\ SemJRplyLSnonEmpty 
 /\ LSnonEmpty  
 /\ OneJRply  
 /\ SemDead
 /\ SemPRply
 /\ SemProbe
 /\ PrbMsgDual
 /\ OnePRply
 /\ OneProbe
 /\ SemProbePRply  
 /\ SemLease
 \*old
 /\ HalfNeighbor
 /\ LeaseGrant  
 /\ SemJReq 
 /\ SemJoinNeighbor
 /\ OneJReq
 /\ JoinMsgDual 
\* general invariants
 \*/\ NeighborClosest
 \* critical invariants
 \*/\ CompleteLeafSet
 \*/\ CompleteLeafSetL
 \*/\ ReadyNeighborRight
 \*/\ ReadyNeighborLeft
 \*/\ NoCCJoin
 \*/\ NoCCJoinL
 \* lemmas
 \*/\ LSetInv 
 \*/\ GrantNeighbor
 \*/\ CoverToJ
 \*/\ SingleReadyNodeToj
 \*/\ SingleReadyNode
 \*/\ SemSingleReadyJoin
 \*/\ SemEmptyLSetSingleNode
 \*/\ SemJReq


 
 \*/\ AtleastOneReady
\* lookup msg is always delivered by the closest "ready" node to its key
Correctness == 
[][\A i, k, n \in I: Deliver(i,k) /\ status[n] = "ready" 
         => AbsDist(i, k) =< AbsDist(n, k)]_vars
                     
Consistency == \* At any time, there exist at most one node taking charge of a key k
[][\A x, y, k \in I: ENABLED Deliver(x, k) /\ ENABLED Deliver(y, k) => x = y]_vars
(*\A x, y, k \in I: 
   (   status[x] = "ready"   
    /\ LeftDist(LeftCover(lset[x]), k) 
       =< LeftDist(LeftCover(lset[x]), RightCover(lset[x]))
    /\ status[y] = "ready"
    /\ LeftDist(LeftCover(lset[y]), k) 
       =< LeftDist(LeftCover(lset[y]), RightCover(lset[y]))
    )
    => x = y
*)
CorrectDelivery == \A i,k \in I: 
  ENABLED Deliver(i, k)
  => /\ \A n \in I: status[n] = "ready"=> AbsDist(i, k) =< AbsDist(n, k)
     /\ \A j \in I\{i}: ~ENABLED Deliver(j, k) 
     
CorrectCoverage == \A i,k \in I: 
  /\ status[i] = "ready" 
  /\ \E m \in receivedMsgs: 
     /\ m.mreq.type = "Lookup"
     /\ m.destination = i
     /\ m.mreq.node = k
     /\ Covers(lset[i], k)
  => /\ \A n \in I: status[n] = "ready"=> AbsDist(i, k) =< AbsDist(n, k)
     /\ \A j \in I\{i}: ~ /\ status[j] = "ready" 
                          /\ \E m \in receivedMsgs: 
                             /\ m.mreq.type = "Lookup"
                             /\ m.destination = j
                             /\ m.mreq.node = k
                             /\ Covers(lset[j], k)
                  
==============================================================
